from typing import Any, Optional

from cffi import FFI

from .arrow_odbc import lib, ffi  # type: ignore
from .buffer import to_bytes_and_len
from .error import raise_on_error


class ConnectionRaii:
    """
    A strong reference to an ODBC connection.

    RAII stands for Resource Acquisition Is Initialization. It is an idiom stemming from the C++
    programming language that ensures resources are properly released when they are no longer
    needed. We use it here to indicate that the sole purpose of this class is to manage the lifetime
    of a strong refrence to an ODBC connection.

    The user of this library should not interact with this class directly, but rather use the
    ``Connection`` class in the ``arrow_odbc.connect`` module.
    """

    def __init__(self, handle: Any) -> None:
        self.handle = handle

    def arrow_odbc_connection(self) -> Any:
        """
        Give access to the inner ArrowOdbcConnection handle
        """
        return self.handle

    def __del__(self):
        if self.handle:
            # Free the resources associated with this handle.
            lib.arrow_odbc_connection_free(self.handle)

    @classmethod
    def connect(
        cls,
        connection_string: str,
        user: Optional[str] = None,
        password: Optional[str] = None,
        login_timeout_sec: Optional[int] = None,
        packet_size: Optional[int] = None,
    ) -> "ConnectionRaii":
        connection_string_bytes = connection_string.encode("utf-8")

        (user_bytes, user_len) = to_bytes_and_len(user)
        (password_bytes, password_len) = to_bytes_and_len(password)
        # We use a pointer to pass the login time, so NULL can represent None
        if login_timeout_sec is None:
            login_timeout_sec_ptr = FFI.NULL
        else:
            login_timeout_sec_ptr = ffi.new("uint32_t *")
            login_timeout_sec_ptr[0] = login_timeout_sec
        if packet_size is None:
            packet_size_ptr = FFI.NULL
        else:
            packet_size_ptr = ffi.new("uint32_t *")
            packet_size_ptr[0] = packet_size
        connection_out = ffi.new("ArrowOdbcConnection **")
        # Open connection to ODBC Data Source
        error = lib.arrow_odbc_connection_make(
            connection_string_bytes,
            len(connection_string_bytes),
            user_bytes,
            user_len,
            password_bytes,
            password_len,
            login_timeout_sec_ptr,
            packet_size_ptr,
            connection_out,
        )
        # See if we connected successfully and return an error if not
        raise_on_error(error)
        # Dereference output pointer. This gives us an `ArrowOdbcConnection *`. We take ownership of
        # the ArrowOdbcConnection and must take care to free it.
        handle = connection_out[0]
        return ConnectionRaii(handle=handle)



from cffi import FFI


def to_bytes_and_len(value: str | None) -> tuple[FFI.CType | bytes, int]:
    if value is None:
        value_bytes = FFI.NULL
        value_len = 0
    else:
        value_bytes = value.encode("utf-8")
        value_len = len(value_bytes)

    return (value_bytes, value_len)



# pyright: reportAttributeAccessIssue=false
# pyright: reportMissingTypeStubs=false
# pyright: reportIndexIssue=false

from collections.abc import Callable, Sequence
from enum import Enum

import pyarrow
from cffi.api import FFI  # type: ignore
from pyarrow import Array, RecordBatch, Schema  # type: ignore
from pyarrow.cffi import ffi as arrow_ffi  # type: ignore

from .arrow_odbc import ffi, lib  # type: ignore
from .buffer import to_bytes_and_len
from .connection_raii import ConnectionRaii
from .error import raise_on_error

# Default maximum buffer size for transition buffer. Defaults to 512 MiB.
DEFAULT_FETCH_BUFFER_LIMIT_IN_BYTES = 2**29

# Default maximum row size for transition buffer. Defaults to 65535 rows. This is the maximum row
# size which can be stored in a signed 16 bit integer. This is a reasonably large upper bound for
# batch sizes, yet avoids trouble with ODBC drivers, which use signed 16 bit integers to represent
# some of the indices internally.
DEFAULT_FETCH_BUFFER_LIMIT_IN_ROWS = 65535


class TextEncoding(Enum):
    """
    Text encoding used for the payload of text columns, to transfer data from the data source to the
    application.

    ``Auto`` evaluates to Utf16 on windows and Self::Utf8 on other systems. We do this, because most
    systems e.g. MacOs and Linux use UTF-8 as their default encoding, while windows may still use a
    Latin1 or some other extended ASCII as their narrow encoding. On the other hand many Posix
    drivers are lacking in their support for wide function calls and UTF-16. So using ``Utf16`` on
    windows and ``Utf8`` everythere else is a good starting point.

    ``Utf8`` use narrow characters (one byte) to encode text in payloads. ODBC lets the client
    choose the encoding which should be based on the system local. This is often not what is
    actually happening though. If we use narrow encoding, we assume the text to be UTF-8 and error
    if we find that not to be the case.

    ``Utf16`` use wide characters (two bytes) to encode text in payloads. ODBC defines the encoding
    to be always UTF-16.
    """

    AUTO = 0
    UTF8 = 1
    UTF16 = 2


def _schema_from_handle(handle) -> Schema:
    """
    Take a handle to an ArrowOdbcReader and return the associated pyarrow schema
    """
    # Expose schema as attribute
    # https://github.com/apache/arrow/blob/5ead37593472c42f61c76396dde7dcb8954bde70/python/pyarrow/tests/test_cffi.py
    with arrow_ffi.new("struct ArrowSchema *") as schema_out:
        error = lib.arrow_odbc_reader_schema(handle, schema_out)
        raise_on_error(error)
        ptr_schema = int(ffi.cast("uintptr_t", schema_out))
        return Schema._import_from_c(ptr_schema)


class _BatchReaderRaii:
    """
    Takes ownership of the reader in its various states and makes sure its resources are freed if
    the object is deleted.
    """

    def __init__(self):
        reader_out = ffi.new("ArrowOdbcReader **")
        lib.arrow_odbc_reader_make(reader_out)
        # We take ownership of the corresponding reader written in Rust and keep it alive until
        # `self` is deleted.
        self.handle = reader_out[0]

    def __del__(self):
        # Free the resources associated with this handle.
        lib.arrow_odbc_reader_free(self.handle)

    def schema(self) -> Schema:
        return _schema_from_handle(self.handle)

    def next_batch(self):
        array = arrow_ffi.new("struct ArrowArray *")
        schema = arrow_ffi.new("struct ArrowSchema *")

        has_next_out = ffi.new("int*")

        error = lib.arrow_odbc_reader_next(self.handle, array, schema, has_next_out)
        raise_on_error(error)

        if has_next_out[0] == 0:
            return None
        else:
            array_ptr = int(ffi.cast("uintptr_t", array))
            schema_ptr = int(ffi.cast("uintptr_t", schema))
            struct_array = Array._import_from_c(array_ptr, schema_ptr)
            return RecordBatch.from_struct_array(struct_array)

    def query(
        self,
        connection: ConnectionRaii,
        query: str,
        parameters: Sequence[str | None] | None,
        text_encoding: TextEncoding,
        query_timeout_sec: int | None,
    ):
        query_bytes = query.encode("utf-8")

        if parameters is None:
            parameters_array = FFI.NULL
            parameters_len = 0
            encoded_parameters = []
        else:
            # Check precondition in order to save users some debugging, in case they directly pass a
            # non-string argument and do not use a type linter.
            if not all([p is None or hasattr(p, "encode") for p in parameters]):
                raise TypeError(
                    "read_arrow_batches_from_odbc only supports string arguments for SQL query "
                    "parameters"
                )

            parameters_array = ffi.new("ArrowOdbcParameter *[]", len(parameters))
            parameters_len = len(parameters)
            # Must be kept alive. Within Rust code we only allocate an additional indicator the
            # string payload is just referenced.
            encoded_parameters = [to_bytes_and_len(p) for p in parameters]

        text_encoding = text_encoding.value

        for p_index in range(0, parameters_len):
            (p_bytes, p_len) = encoded_parameters[p_index]
            parameters_array[p_index] = lib.arrow_odbc_parameter_string_make(
                p_bytes, p_len, text_encoding
            )

        if query_timeout_sec is None:
            query_timeout_sec_pointer = ffi.NULL
        else:
            query_timeout_sec_pointer = ffi.new("uintptr_t *")
            query_timeout_sec_pointer[0] = query_timeout_sec

        error = lib.arrow_odbc_reader_query(
            self.handle,
            connection.arrow_odbc_connection(),
            query_bytes,
            len(query_bytes),
            parameters_array,
            parameters_len,
            query_timeout_sec_pointer,
        )

        # See if we managed to execute the query successfully and return an error if not
        raise_on_error(error)

    def bind_buffers(
        self,
        batch_size: int,
        max_bytes_per_batch: int,
        max_text_size: int,
        max_binary_size: int,
        falliable_allocations: bool,
        payload_text_encoding: TextEncoding,
        schema: Schema | None,
        map_schema: Callable[[Schema], Schema] | None,
        fetch_concurrently: bool,
    ):
        if map_schema is not None:
            schema = map_schema(self.schema())

        payload_text_encoding_int = payload_text_encoding.value

        ptr_schema = _export_schema_to_c(schema)

        error = lib.arrow_odbc_reader_bind_buffers(
            self.handle,
            batch_size,
            max_bytes_per_batch,
            max_text_size,
            max_binary_size,
            falliable_allocations,
            fetch_concurrently,
            payload_text_encoding_int,
            ptr_schema,
        )
        # See if we managed to execute the query successfully and return an error if not
        raise_on_error(error)

    def more_results(
        self,
    ) -> bool:
        with ffi.new("bool *") as has_more_results_c:
            error = lib.arrow_odbc_reader_more_results(
                self.handle,
                has_more_results_c,
            )
            # See if we managed to execute the query successfully and return an error if not
            raise_on_error(error)
            # Remember wether there is a new result set in a boolean
            has_more_results = has_more_results_c[0] != 0

        return has_more_results


class BatchReader:
    """
    Iterates over Arrow batches from an ODBC data source
    """

    def __init__(self, reader: _BatchReaderRaii):
        """
        Low level constructor, users should rather invoke `read_arrow_batches_from_odbc` in order to
        create instances of `BatchReader`.
        """
        # We take ownership of the corresponding reader written in Rust and keep it alive until
        # `self` is deleted. We also take care to keep this reader either in empty or reader state,
        # meaning we always have it ready to produce batches, or we consumed everything. We avoid
        # exposing the intermediate cursor state directly to users.
        self.reader = reader

        # This is the schema of the batches returned by reader. We take care to keep it in sync in
        # case the state of reader changes.
        self.schema = self.reader.schema()

    @classmethod
    def _from_connection(
        cls,
        connection: ConnectionRaii,
        query: str,
        batch_size: int = DEFAULT_FETCH_BUFFER_LIMIT_IN_ROWS,
        parameters: Sequence[str | None] | None = None,
        max_bytes_per_batch: int | None = DEFAULT_FETCH_BUFFER_LIMIT_IN_BYTES,
        max_text_size: int | None = None,
        max_binary_size: int | None = None,
        falliable_allocations: bool = False,
        schema: Schema | None = None,
        map_schema: Callable[[Schema], Schema] | None = None,
        fetch_concurrently=True,
        query_timeout_sec: int | None = None,
        payload_text_encoding: TextEncoding = TextEncoding.AUTO,
    ) -> "BatchReader":
        reader = _BatchReaderRaii()

        reader.query(
            connection=connection,
            query=query,
            parameters=parameters,
            text_encoding=payload_text_encoding,
            query_timeout_sec=query_timeout_sec,
        )

        if max_text_size is None:
            max_text_size = 0
        if max_binary_size is None:
            max_binary_size = 0
        if max_bytes_per_batch is None:
            max_bytes_per_batch = 0

        # Let us transition to reader state
        reader.bind_buffers(
            batch_size=batch_size,
            max_bytes_per_batch=max_bytes_per_batch,
            max_text_size=max_text_size,
            max_binary_size=max_binary_size,
            falliable_allocations=falliable_allocations,
            payload_text_encoding=payload_text_encoding,
            schema=schema,
            map_schema=map_schema,
            fetch_concurrently=fetch_concurrently,
        )

        return BatchReader(reader)

    def __iter__(self):
        # Implement iterable protocol so reader can be used in for loops.
        return self

    def __next__(self) -> RecordBatch:
        # Implement iterator protocol
        batch = self.reader.next_batch()
        if batch is None:
            raise StopIteration()
        else:
            return batch

    def more_results(
        self,
        batch_size: int = DEFAULT_FETCH_BUFFER_LIMIT_IN_ROWS,
        max_bytes_per_batch: int = DEFAULT_FETCH_BUFFER_LIMIT_IN_BYTES,
        max_text_size: int | None = None,
        max_binary_size: int | None = None,
        falliable_allocations: bool = False,
        schema: Schema | None = None,
        map_schema: Callable[[Schema], Schema] | None = None,
        fetch_concurrently=True,
        payload_text_encoding: TextEncoding = TextEncoding.AUTO,
    ) -> bool:
        """
        Move the reader to the next result set returned by the data source.

        A datasource may return multiple results if multiple SQL statements are executed in a single
        query or a stored procedure is called. This method closes the current cursor and moves it to
        the next result set. You may move to the next result set without extracting the current one.

        Example:

        .. code-block:: python

            from arrow_odbc import read_arrow_batches_from_odbc

            connection_string=
                "Driver={ODBC Driver 18 for SQL Server};" \
                "Server=localhost;" \
                "TrustServerCertificate=yes;"

            reader = read_arrow_batches_from_odbc(
                query=f"SELECT * FROM MyTable; SELECT * FROM OtherTable;",
                connection_string=connection_string,
                batch_size=1000,
                user="SA",
                password="My@Test@Password",
            )

            # Process first result
            for batch in reader:
                # Process arrow batches
                df = batch.to_pandas()
                # ...

            reader.more_results()

            # Process second result
            for batch in reader:
                # Process arrow batches
                df = batch.to_pandas()
                # ...

        :param batch_size: The maximum number rows within each batch. Batch size can be individually
            choosen for each result set. The maximum number of rows can be less if the upper bound
            defined by ``max_bytes_per_batch`` is lower.
        :param max_bytes_per_batch: An upper limit for the total size (all columns) of the buffer
            used to transit data from the ODBC driver to the application. Please note that memory
            consumption of this buffer is determined not by the actual values, but by the maximum
            possible length of an indiviual row times the number of rows it can hold. Both
            ``batch_size`` and this parameter define upper bounds for the same buffer. Which ever
            bound is lower is used to determine the buffer size.
        :param max_text_size: In order for fast bulk fetching to work, `arrow-odbc` needs to know the
            size of the largest possible field in each column. It will do so itself automatically by
            considering the schema information. However, trouble arises if the schema contains
            unbounded variadic fields like `VARCHAR(MAX)` which can hold really large values. These
            have a very high upper element size, if any. In order to work with such schemas we need
            a limit, of what the an upper bound of the actual values in the column is, as opposed to
            the what the largest value is the column could theoretically store. There is no need for
            this to be precise, but just knowing that a value would never exceed 4KiB rather than
            2GiB is enough to allow for tremendous efficiency gains. The size of the text is
            specified in UTF-8 encoded bytes if using a narrow encoding (typically all non-windows
            systems) and in UTF-16 encoded pairs of bytes on systems using a wide encoding
            (typically windows). This means about the size in letters, yet if you are using a lot of
            emojis or other special characters this number might need to be larger.
        :param max_binary_size: An upper limit for the size of buffers bound to variadic binary
            columns of the data source. This limit does not (directly) apply to the size of the
            created arrow buffers, but rather applies to the buffers used for the data in transit.
            Use this option if you have e.g. VARBINARY(MAX) fields in your next batch.
        :param falliable_allocations: If ``True`` an recoverable error is raised in case there is
            not enough memory to allocate the buffers. This option may incurr a performance penalty
            which scales with the batch size parameter (but not with the amount of actual data in
            the source). In case you can test your query against the schema you can safely set this
            to ``False``. The required memory will not depend on the amount of data in the data
            source. Default is ``True`` though, safety first.
        :param fetch_concurrently: Trade memory for speed. Allocates another transit buffer and use
            it to fetch row set groups (aka. batches) from the ODBC data source in a dedicated
            system thread, while the main thread converts the previous batch to arrow arrays and
            executes the application logic. The transit buffer may be the biggest part of the
            required memory so if ``True`` ``arrow-odbc`` consumes almost two times the memory as
            compared to false. On the flipsite the next batch can be fetched from the database
            immediatly without waiting for the application logic to return control.
        :param payload_text_encoding: Controls the encoding used for transferring text data from the
            ODBC data source to the application. The resulting Arrow arrays will still be UTF-8
            encoded. You may want to use this if you get garbage characters or invalid UTF-8 errors
            on non-windows systems to set the encoding to ``TextEncoding.Utf16``. On windows systems
            you may want to set this to ``TextEncoding::Utf8`` to gain performance benefits, after
            you have verified that your system locale is set to UTF-8.
        :return: ``True`` in case there is another result set. ``False`` in case that the last
            result set has been processed.
        """
        if max_text_size is None:
            max_text_size = 0
        if max_binary_size is None:
            max_binary_size = 0

        has_more_results = self.reader.more_results()

        self.reader.bind_buffers(
            batch_size=batch_size,
            max_bytes_per_batch=max_bytes_per_batch,
            max_text_size=max_text_size,
            max_binary_size=max_binary_size,
            falliable_allocations=falliable_allocations,
            payload_text_encoding=payload_text_encoding,
            schema=schema,
            map_schema=map_schema,
            fetch_concurrently=fetch_concurrently,
        )

        # Every result set can have its own schema, so we must update our member
        self.schema = self.reader.schema()

        return has_more_results

    def into_pyarrow_record_batch_reader(self):
        """
        Converts the ``arrow-odbc`` ``BatchReader`` into a ``pyarrow`` ``RecordBatchReader``. This
        method fully passes ownership to the new reader and leaves ``self`` empty.

        ``arrow-odbc``s BatchReader interface offers some functionality specific to ODBC
        datasources. E.g. the ability to move to the next result set of a stored procedure. You may
        not need this extra functionality and would rather like to integrate the ``BatchReader``
        with other libraries like e.g. DuckDB. In order to do this you can use this method to
        convert the ``arrow-odbc`` BatchReader into a ``pyarrow`` ``RecordBatchReader``.
        """
        # New empty tmp reader
        reader = _BatchReaderRaii()
        tmp = BatchReader(reader)
        # Swap self and tmp
        tmp.reader, self.reader = self.reader, tmp.reader
        tmp.schema, self.schema = self.schema, tmp.schema
        return pyarrow.RecordBatchReader.from_batches(tmp.schema, tmp)


def _export_schema_to_c(schema):
    if schema is None:
        ptr_schema = ffi.NULL
    else:
        ptr_schema = arrow_ffi.new("struct ArrowSchema *")
        int_schema = int(ffi.cast("uintptr_t", ptr_schema))
        schema._export_to_c(int_schema)
    return ptr_schema



from typing import Any, Callable, Sequence

from pyarrow import Schema  # type: ignore
from pyarrow import RecordBatchReader

from .reader import (
    DEFAULT_FETCH_BUFFER_LIMIT_IN_BYTES,
    DEFAULT_FETCH_BUFFER_LIMIT_IN_ROWS,
    BatchReader,
    TextEncoding,
)
from .connection_raii import ConnectionRaii
from .pool import enable_odbc_connection_pooling
from .writer import BatchWriter


class Connection:
    """
    A strong reference to an ODBC connection.
    """

    def __init__(self, raii: ConnectionRaii) -> None:
        self.raii = raii

    @classmethod
    def enable_connection_pooling(cls) -> None:
        """
        Activates the connection pooling of the ODBC driver manager for the entire process. Best
        called before creating the ODBC environment, i.e. before you the first connection is opend
        with arrow-odbc. This is useful in scenarios there you frequently read or write rows and the
        overhead of creating a connection for each query is significant.

        Example:

        .. code-block:: python

            from arrow_odbc import Connection

            # Let the ODBC driver manager take care of connection pooling for us
            Connection.enable_connection_pooling()

            # Create the first connection after Connection pooling is enabled
            connection_string=
                "Driver={ODBC Driver 18 for SQL Server};" \
                "Server=localhost;" \
                "TrustServerCertificate=yes;"
            connection = connect(
                connection_string=connection_string,
                user="SA",
                password="My@Test@Password"
            )
        """
        enable_odbc_connection_pooling()

    def read_arrow_batches(
        self,
        query: str,
        batch_size: int = DEFAULT_FETCH_BUFFER_LIMIT_IN_ROWS,
        parameters: Sequence[str | None] | None = None,
        max_bytes_per_batch: int | None = DEFAULT_FETCH_BUFFER_LIMIT_IN_BYTES,
        max_text_size: int | None = None,
        max_binary_size: int | None = None,
        falliable_allocations: bool = False,
        schema: Schema | None = None,
        map_schema: Callable[[Schema], Schema] | None = None,
        fetch_concurrently: bool = True,
        query_timeout_sec: int | None = None,
        payload_text_encoding: TextEncoding = TextEncoding.AUTO,
    ) -> BatchReader:
        """
        Execute the query and read the result as an iterator over Arrow batches.

        Example:

        .. code-block:: python

            from arrow_odbc import connect

            # Connect to the data source
            connection_string=
                "Driver={ODBC Driver 18 for SQL Server};" \
                "Server=localhost;" \
                "TrustServerCertificate=yes;"

            connection = connect(
                connection_string=connection_string,
                user="SA",
                password="My@Test@Password"
            )

            # Execute query and create reader
            reader = connection.read_arrow_batches(
                query=f"SELECT * FROM MyTable WHERE a=?",
                batch_size=1000,
                parameters=["I'm a positional query parameter"],
            )

            # Process results
            for batch in reader:
                # Process arrow batches
                df = batch.to_pandas()
                # ...

        :param connection: An ODBC connection created with ``connect``.
        :param query: The SQL statement yielding the result set which is converted into arrow record
            batches.
        :param batch_size: The maximum number rows within each batch. The maximum number of rows can
            be less if the upper bound defined by ``max_bytes_per_batch`` is lower.
        :param parameters: ODBC allows you to use a question mark as placeholder marker (``?``) for
            positional parameters. This argument takes a list of parameters those number must match
            the number of placholders in the SQL statement. Using this instead of literals helps you
            avoid SQL injections or may otherwise simplify your code. Currently all parameters are
            passed as VARCHAR strings. You can use `None` to pass `NULL`.
        :param max_bytes_per_batch: An upper limit for the total size (all columns) of the buffer
            used to transit data from the ODBC driver to the application. Please note that memory
            consumption of this buffer is determined not by the actual values, but by the maximum
            possible length of an individual row times the number of rows it can hold. Both
            ``batch_size`` and this parameter define upper bounds for the same buffer. Which ever
            bound is lower is used to determine the buffer size.
        :param max_text_size: In order for fast bulk fetching to work, `arrow-odbc` needs to know
            the size of the largest possible field in each column. It will do so itself
            automatically by considering the schema information. However, trouble arises if the
            schema contains unbounded variadic fields like `VARCHAR(MAX)` which can hold really
            large values. These have a very high upper element size, if any. In order to work with
            such schemas we need a limit, of what the an upper bound of the actual values in the
            column is, as opposed to the what the largest value is the column could theoretically
            store. There is no need for this to be precise, but just knowing that a value would
            never exceed 4KiB rather than 2GiB is enough to allow for tremendous efficiency gains.
            The size of the text is specified in UTF-8 encoded bytes if using a narrow encoding
            (typically all non-windows systems) and in UTF-16 encoded pairs of bytes on systems
            using a wide encoding (typically windows). This means about the size in letters, yet if
            you are using a lot of emojis or other special characters this number might need to be
            larger.
        :param max_binary_size: An upper limit for the size of buffers bound to variadic binary
            columns of the data source. This limit does not (directly) apply to the size of the
            created arrow buffers, but rather applies to the buffers used for the data in transit.
            Use this option if you have e.g. VARBINARY(MAX) fields in your database schema. In such
            a case without an upper limit, the ODBC driver of your data source is asked for the
            maximum size of an element, and is likely to answer with either 0 or a value which is
            way larger than any actual entry in the column. If you can not adapt your database
            schema, this limit might be what you are looking for. This is the maximum size in bytes
            of the binary column.
        :param falliable_allocations: If ``True`` an recoverable error is raised in case there is
            not enough memory to allocate the buffers. This option may incurr a performance penalty
            which scales with the batch size parameter (but not with the amount of actual data in
            the source). In case you can test your query against the schema you can safely set this
            to ``False``. The required memory will not depend on the amount of data in the data
            source. Default is ``True`` though, safety first.
        :param schema: Allows you to overwrite the automatically detected schema with one supplied
            by the application. Reasons for doing so include domain knowledge you have about the
            data which is not reflected in the schema information. E.g. you happen to know a field
            of timestamps contains strictly dates. Another reason could be that for certain usecases
            another it can make sense to decide the type based on what you want to do with it,
            rather than its source. E.g. if you simply want to put everything into a CSV file it can
            make perfect sense to fetch everything as string independent of its source type.
        :param map_schema: Allows you to provide a custom schema based on the schema inferred from
            the metainformation of the query. This would allow you to e.g. map every column type to
            string or replace any float32 with a float64, or anything else you might want to
            customize, for various reasons while still staying generic over the input schema. If
            both ``map_schema`` and ``schema`` are specified ``map_schema`` takes priority.
        :param fetch_concurrently: Trade memory for speed. Allocates another transit buffer and use
            it to fetch row set groups (aka. batches) from the ODBC data source in a dedicated
            system thread, while the main thread converts the previous batch to arrow arrays and
            executes the application logic. The transit buffer may be the biggest part of the
            required memory so if ``True`` ``arrow-odbc`` consumes almost two times the memory as
            compared to false. On the flipsite the next batch can be fetched from the database
            immediatly without waiting for the application logic to return control.
        :param query_timeout_sec: Use this to limit the time the query is allowed to take, before
            responding with data to the application. The driver may replace the number of seconds
            you provide with a minimum or maximum value. You can specify ``0``, to deactivate the
            timeout, this is the default. For this to work the driver must support this feature.
            E.g. PostgreSQL, and Microsoft SQL Server do, but SQLite or MariaDB do not.
        :param payload_text_encoding: Controls the encoding used for transferring text data from the
                ODBC data source to the application. The resulting Arrow arrays will still be UTF-8
                encoded. You may want to use this if you get garbage characters or invalid UTF-8
                errors on non-windows systems to set the encoding to ``TextEncoding.Utf16``. On
                windows systems you may want to set this to ``TextEncoding::Utf8`` to gain
                performance benefits, after you have verified that your system locale is set to
                UTF-8.
        :return: A ``BatchReader`` is returned, which implements the iterator protocol and iterates
            over individual arrow batches.
        """
        return BatchReader._from_connection(
            connection=self.raii,
            query=query,
            batch_size=batch_size,
            parameters=parameters,
            max_bytes_per_batch=max_bytes_per_batch,
            max_text_size=max_text_size,
            max_binary_size=max_binary_size,
            falliable_allocations=falliable_allocations,
            schema=schema,
            map_schema=map_schema,
            fetch_concurrently=fetch_concurrently,
            query_timeout_sec=query_timeout_sec,
            payload_text_encoding=payload_text_encoding,
        )

    def insert_into_table(
        self,
        reader: Any,
        table: str,
        chunk_size: int,
    ):
        """
        Consume the batches in the reader and insert them into a table on the database.

        Example:

        .. code-block:: python

            from arrow_odbc import connect
            import pyarrow as pa
            import pandas


            def dataframe_to_table(df):
                table = pa.Table.from_pandas(df)
                reader = pa.RecordBatchReader.from_batches(table.schema, table.to_batches())
                connectiion = connect(
                    connection_string=connection_string,
                    user="SA",
                    password="My@Test@Password"
                )
                connection.insert_into_table(
                    table="MyTable",
                    reader=reader,
                    chunk_size=1000,
                )

        :param reader: Reader is used to iterate over record batches. It must expose a `schema`
            attribute, referencing an Arrow schema. Each field in the schema must correspond to a
            column in the table with identical name. The iterator must yield individual arrow tables
        :param table: Name of a database table to insert into. Used to generate the insert statement
            for the bulk writer.
        :param chunk_size: Number of records to insert in each roundtrip to the database.
            Independent of batch size (i.e. number of rows in an individual record batch).
        """
        writer = BatchWriter._from_connection(
            connection=self.raii,
            reader=reader,
            chunk_size=chunk_size,
            table=table,
        )

        # Write all batches in reader
        for batch in reader:
            writer.write_batch(batch)
        writer.flush()

    def from_table_to_db(
        self,
        source: Any,
        target: str,
        chunk_size: int = 1000,
    ):
        """
        Reads an arrow table and inserts its contents into a relational table on the database.

        This is a convinience wrapper around ``insert_into_table`` which converts an arrow table
        into a record batch reader for you.

        Example:

        .. code-block:: python

            from arrow_odbc import connect
            import pyarrow as pa
            import pandas


            def dataframe_to_table(df):
                table = pa.Table.from_pandas(df)
                connection = connect(
                    connection_string=connection_string,
                    user="SA",
                    password="My@Test@Password"
                )
                connection.from_table_to_db(
                    source=table,
                    target="MyTable",
                    chunk_size=1000
                )

        :param source: PyArrow table with content to be inserted into the target table on the
            database. Each column of the table must correspond to a column in the target table with
            identical name.
        :param target: Name of the database table to insert into.
        :param chunk_size: Number of records to insert in each roundtrip to the database. The number
            will be automatically reduced to the number of rows, if the table is small, in order to
            save memory.
        """
        # There is no need for chunk size to exceed the maximum amount of rows in the table
        chunk_size = min(chunk_size, source.num_rows)
        # We implemement this in terms of the functionality to insert a batches from a record batch
        # reader, so first we convert our table into a record batch reader.
        schema = source.schema
        batches = source.to_batches(chunk_size)
        reader = RecordBatchReader.from_batches(schema, batches)
        # Now we can insert from the reader
        self.insert_into_table(
            reader=reader,
            table=target,
            chunk_size=chunk_size,
        )


def connect(
    connection_string: str,
    user: str | None = None,
    password: str | None = None,
    login_timeout_sec: int | None = None,
    packet_size: int | None = None,
) -> Connection:
    """
    Opens a connection to an ODBC data source.

    In case you want to use connection pooling, call ``enable_odbc_connection_pooling()`` before
    calling this function.

    Example:

    .. code-block:: python

        from arrow_odbc import connect, enable_odbc_connection_pooling

        # Let the ODBC driver manager take care of connection pooling for us
        enable_odbc_connection_pooling()

        # Connect to the data source
        connection_string=
            "Driver={ODBC Driver 18 for SQL Server};" \
            "Server=localhost;" \
            "TrustServerCertificate=yes;"
        connection = connect(
            connection_string=connection_string,
            user="SA",
            password="My@Test@Password"
        )

    :param connection_string: ODBC Connection string used to connect to the data source. To find a
        connection string for your data source try https://www.connectionstrings.com/.
    :param user: Allows for specifying the user seperatly from the connection string if it is not
        already part of it. The value will eventually be escaped and attached to the connection
        string as `UID`.
    :param password: Allows for specifying the password seperatly from the connection string if it
        is not already part of it. The value will eventually be escaped and attached to the
        connection string as `PWD`.
    :param login_timeout_sec: Number of seconds to wait for a login request to complete before
        returning to the application. The default is driver-dependent. If ``0``, the timeout is
        disabled and a connection attempt will wait indefinitely. If the specified timeout exceeds
        the maximum login timeout in the data source, the driver substitutes that value and uses
        that instead.
    :param packet_size: Specifying the network packet size in bytes. Many ODBC drivers do not
        support this option. If the specified size exceeds the maximum packet size or is smaller
        than the minimum packet size, the driver substitutes that value and returns SQLSTATE 01S02
        (Option value changed).You may want to enable logging to standard error using
        ``log_to_stderr``.
    :return: A ``Connection`` is returned.
    """
    raii = ConnectionRaii.connect(
        connection_string=connection_string,
        user=user,
        password=password,
        login_timeout_sec=login_timeout_sec,
        packet_size=packet_size,
    )
    return Connection(raii=raii)



from typing import Any

from pyarrow.cffi import ffi as arrow_ffi

from .arrow_odbc import ffi, lib  # type: ignore
from .connection_raii import ConnectionRaii
from .error import raise_on_error


class BatchWriter:
    """
    Writes arrow batches to a database table.
    """

    def __init__(
        self,
        handle,
    ):
        # We take ownership of the corresponding writer written in Rust and keep it alive until
        # `self` is deleted
        self.handle = handle

    def __del__(self):
        # Free the resources associated with this handle.
        lib.arrow_odbc_writer_free(self.handle)

    @classmethod
    def _from_connection(
        cls,
        connection: ConnectionRaii,
        reader: Any,
        chunk_size: int,
        table: str,
    ):
        table_bytes = table.encode("utf-8")

        # Allocate structures where we will export the Array data and the Array schema. They will be
        # released when we exit the with block.
        with arrow_ffi.new("struct ArrowSchema*") as c_schema:
            # Get the references to the C Data structures.
            c_schema_ptr = int(arrow_ffi.cast("uintptr_t", c_schema))

            # Export the schema to the C Data structures.
            reader.schema._export_to_c(c_schema_ptr)

            writer_out = ffi.new("ArrowOdbcWriter **")
            error = lib.arrow_odbc_writer_make(
                connection.arrow_odbc_connection(),
                table_bytes,
                len(table_bytes),
                chunk_size,
                c_schema,
                writer_out,
            )
            raise_on_error(error)

        return BatchWriter(handle=writer_out[0])

    def write_batch(self, batch):
        """
        Fills the internal buffers of the writer with data from the batch. Every
        time they are full, the data is send to the database. To make sure all
        the data is is send ``flush`` must be called.
        """
        with (
            arrow_ffi.new("struct ArrowArray*") as c_array,
            arrow_ffi.new("struct ArrowSchema*") as c_schema,
        ):
            # Get the references to the C Data structures
            c_array_ptr = int(arrow_ffi.cast("uintptr_t", c_array))
            c_schema_ptr = int(arrow_ffi.cast("uintptr_t", c_schema))

            # Export the Array to the C Data structures.
            batch._export_to_c(c_array_ptr)
            batch.schema._export_to_c(c_schema_ptr)

            error = lib.arrow_odbc_writer_write_batch(self.handle, c_array, c_schema)
            raise_on_error(error)

    def flush(self):
        """
        Inserts the remaining rows of the last chunk to the database.
        """
        error = lib.arrow_odbc_writer_flush(self.handle)
        raise_on_error(error)



from typing import Optional, Callable, Sequence

from pyarrow import Schema

from .connect import connect
from .reader import (
    DEFAULT_FETCH_BUFFER_LIMIT_IN_BYTES,
    DEFAULT_FETCH_BUFFER_LIMIT_IN_ROWS,
    BatchReader,
    TextEncoding,
)


def read_arrow_batches_from_odbc(
    query: str,
    connection_string: str,
    batch_size: int = DEFAULT_FETCH_BUFFER_LIMIT_IN_ROWS,
    user: Optional[str] = None,
    password: Optional[str] = None,
    parameters: Optional[Sequence[Optional[str]]] = None,
    max_bytes_per_batch: Optional[int] = DEFAULT_FETCH_BUFFER_LIMIT_IN_BYTES,
    max_text_size: Optional[int] = None,
    max_binary_size: Optional[int] = None,
    falliable_allocations: bool = False,
    login_timeout_sec: Optional[int] = None,
    packet_size: Optional[int] = None,
    schema: Optional[Schema] = None,
    map_schema: Optional[Callable[[Schema], Schema]] = None,
    fetch_concurrently=True,
    query_timeout_sec: Optional[int] = None,
    payload_text_encoding: TextEncoding = TextEncoding.AUTO,
) -> BatchReader:
    """
    Execute the query and read the result as an iterator over Arrow batches.

    Example:

    .. code-block:: python

        from arrow_odbc import read_arrow_batches_from_odbc

        connection_string=
            "Driver={ODBC Driver 18 for SQL Server};" \
            "Server=localhost;"
            "TrustServerCertificate=yes;"

        reader = read_arrow_batches_from_odbc(
            query=f"SELECT * FROM MyTable WHERE a=?",
            connection_string=connection_string,
            batch_size=1000,
            parameters=["I'm a positional query parameter"],
            user="SA",
            password="My@Test@Password",
        )

        for batch in reader:
            # Process arrow batches
            df = batch.to_pandas()
            # ...

    :param query: The SQL statement yielding the result set which is converted into arrow record
        batches.
    :param batch_size: The maximum number rows within each batch. The maximum number of rows can be
        less if the upper bound defined by ``max_bytes_per_batch`` is lower.
    :param connection_string: ODBC Connection string used to connect to the data source. To find a
        connection string for your data source try https://www.connectionstrings.com/.
    :param user: Allows for specifying the user seperatly from the connection string if it is not
        already part of it. The value will eventually be escaped and attached to the connection
        string as `UID`.
    :param password: Allows for specifying the password seperatly from the connection string if it
        is not already part of it. The value will eventually be escaped and attached to the
        connection string as `PWD`.
    :param parameters: ODBC allows you to use a question mark as placeholder marker (``?``) for
        positional parameters. This argument takes a list of parameters those number must match the
        number of placholders in the SQL statement. Using this instead of literals helps you avoid
        SQL injections or may otherwise simplify your code. Currently all parameters are passed as
        VARCHAR strings. You can use `None` to pass `NULL`.
    :param max_bytes_per_batch: An upper limit for the total size (all columns) of the buffer used
        to transit data from the ODBC driver to the application. Please note that memory consumption
        of this buffer is determined not by the actual values, but by the maximum possible length of
        an indiviual row times the number of rows it can hold. Both ``batch_size`` and this
        parameter define upper bounds for the same buffer. Which ever bound is lower is used to
        determine the buffer size.
    :param max_text_size: In order for fast bulk fetching to work, `arrow-odbc` needs to know the
        size of the largest possible field in each column. It will do so itself automatically by
        considering the schema information. However, trouble arises if the schema contains
        unbounded variadic fields like `VARCHAR(MAX)` which can hold really large values. These have
        a very high upper element size, if any. In order to work with such schemas we need a limit,
        of what the an upper bound of the actual values in the column is, as opposed to the what the
        largest value is the column could theoretically store. There is no need for this to be
        precise, but just knowing that a value would never exceed 4KiB rather than 2GiB is enough to
        allow for tremendous efficiency gains. The size of the text is specified in UTF-8 encoded
        bytes if using a narrow encoding (typically all non-windows systems) and in UTF-16 encoded
        pairs of bytes on systems using a wide encoding (typically windows). This means about the
        size in letters, yet if you are using a lot of emojis or other special characters this
        number might need to be larger.
    :param max_binary_size: An upper limit for the size of buffers bound to variadic binary columns
        of the data source. This limit does not (directly) apply to the size of the created arrow
        buffers, but rather applies to the buffers used for the data in transit. Use this option if
        you have e.g. VARBINARY(MAX) fields in your database schema. In such a case without an upper
        limit, the ODBC driver of your data source is asked for the maximum size of an element, and
        is likely to answer with either 0 or a value which is way larger than any actual entry in
        the column. If you can not adapt your database schema, this limit might be what you are
        looking for. This is the maximum size in bytes of the binary column.
    :param falliable_allocations: If ``True`` an recoverable error is raised in case there is not
        enough memory to allocate the buffers. This option may incurr a performance penalty which
        scales with the batch size parameter (but not with the amount of actual data in the source).
        In case you can test your query against the schema you can safely set this to ``False``. The
        required memory will not depend on the amount of data in the data source. Default is
        ``True`` though, safety first.
    :param login_timeout_sec: Number of seconds to wait for a login request to complete before
        returning to the application. The default is driver-dependent. If ``0``, the timeout is
        disabled and a connection attempt will wait indefinitely. If the specified timeout exceeds
        the maximum login timeout in the data source, the driver substitutes that value and uses
        that instead.
    :param packet_size: Specifying the network packet size in bytes. Many ODBC drivers do not
        support this option. If the specified size exceeds the maximum packet size or is smaller
        than the minimum packet size, the driver substitutes that value and returns SQLSTATE 01S02
        (Option value changed).You may want to enable logging to standard error using
        ``log_to_stderr``.
    :param schema: Allows you to overwrite the automatically detected schema with one supplied by
        the application. Reasons for doing so include domain knowledge you have about the data which
        is not reflected in the schema information. E.g. you happen to know a field of timestamps
        contains strictly dates. Another reason could be that for certain usecases another it can
        make sense to decide the type based on what you want to do with it, rather than its source.
        E.g. if you simply want to put everything into a CSV file it can make perfect sense to fetch
        everything as string independent of its source type.
    :param map_schema: Allows you to provide a custom schema based on the schema inferred from the
        metainformation of the query. This would allow you to e.g. map every column type to string
        or replace any float32 with a float64, or anything else you might want to customize, for
        various reasons while still staying generic over the input schema. If both ``map_schema``
        and ``schema`` are specified ``map_schema`` takes priority.
    :param fetch_concurrently: Trade memory for speed. Allocates another transit buffer and use it
        to fetch row set groups (aka. batches) from the ODBC data source in a dedicated system
        thread, while the main thread converts the previous batch to arrow arrays and executes the
        application logic. The transit buffer may be the biggest part of the required memory so if
        ``True`` ``arrow-odbc`` consumes almost two times the memory as compared to false. On the
        flipsite the next batch can be fetched from the database immediatly without waiting for the
        application logic to return control.
    :param query_timeout_sec: Use this to limit the time the query is allowed to take, before
        responding with data to the application. The driver may replace the number of seconds you
        provide with a minimum or maximum value. You can specify ``0``, to deactivate the timeout,
        this is the default. For this to work the driver must support this feature. E.g. PostgreSQL,
        and Microsoft SQL Server do, but SQLite or MariaDB do not.
    :param payload_text_encoding: Controls the encoding used for transferring text data from the
            ODBC data source to the application. The resulting Arrow arrays will still be UTF-8
            encoded. You may want to use this if you get garbage characters or invalid UTF-8 errors
            on non-windows systems to set the encoding to ``TextEncoding.Utf16``. On windows systems
            you may want to set this to ``TextEncoding::Utf8`` to gain performance benefits, after
            you have verified that your system locale is set to UTF-8.
    :return: A ``BatchReader`` is returned, which implements the iterator protocol and iterates over
        individual arrow batches.
    """
    connection = connect(
        connection_string=connection_string,
        user=user,
        password=password,
        login_timeout_sec=login_timeout_sec,
        packet_size=packet_size,
    )

    return connection.read_arrow_batches(
        query=query,
        batch_size=batch_size,
        parameters=parameters,
        max_bytes_per_batch=max_bytes_per_batch,
        max_text_size=max_text_size,
        max_binary_size=max_binary_size,
        falliable_allocations=falliable_allocations,
        schema=schema,
        map_schema=map_schema,
        fetch_concurrently=fetch_concurrently,
        query_timeout_sec=query_timeout_sec,
        payload_text_encoding=payload_text_encoding,
    )



from .arrow_odbc import lib  # type: ignore
from .error import raise_on_error


def enable_odbc_connection_pooling():
    """
    Activates the connection pooling of the ODBC driver manager for the entire process. Best called
    before creating the ODBC environment, i.e. before you first insert or read rows with arrow-odbc.
    This is useful in scenarios there you frequently read or write rows and the overhead of creating
    a connection for each query is significant.

    Example:

    .. code-block:: python

        from arrow_odbc import connect, enable_odbc_connection_pooling

        # Let the ODBC driver manager take care of connection pooling for us
        enable_odbc_connection_pooling()

        # Connect to the data source
        connection_string=
            "Driver={ODBC Driver 18 for SQL Server};" \
            "Server=localhost;" \
            "TrustServerCertificate=yes;"
        connection = connect(
            connection_string=connection_string,
            user="SA",
            password="My@Test@Password"
        )
    """
    error = lib.arrow_odbc_enable_connection_pooling()
    raise_on_error(error)



from typing import Any

from .connect import connect


def insert_into_table(
    reader: Any,
    chunk_size: int,
    table: str,
    connection_string: str,
    user: str | None = None,
    password: str | None = None,
    login_timeout_sec: int | None = None,
    packet_size: int | None = None,
):
    """
    Consume the batches in the reader and insert them into a table on the database.

    Example:

    .. code-block:: python

        from arrow_odbc import insert_into_table
        import pyarrow as pa
        import pandas


        def dataframe_to_table(df):
            table = pa.Table.from_pandas(df)
            reader = pa.RecordBatchReader.from_batches(table.schema, table.to_batches())
            insert_into_table(
                connection_string=connection_string,
                user="SA",
                password="My@Test@Password",
                chunk_size=1000,
                table="MyTable",
                reader=reader,
            )

    :param reader: Reader is used to iterate over record batches. It must expose a `schema`
        attribute, referencing an Arrow schema. Each field in the schema must correspond to a
        column in the table with identical name. The iterator must yield individual arrow tables. In
        case a column name contains a non-alphanumeric character different from ``@``, ``$``,
        ``#``, or ``_``, the name will be escaped using double quotes. However if the name already
        is already escaped, i.e. it is enclosed in either rectangular brackets ([, ]), double quotes
        (") or backticks (`), it will be used as is. This is to allow for users to apply their own
        quoting if needed, e.g. in case reserved keywords are used as column names.
    :param chunk_size: Number of records to insert in each roundtrip to the database. Independent of
        batch size (i.e. number of rows in an individual record batch).
    :param table: Name of a database table to insert into. Used to generate the insert statement for
        the bulk writer.
    :param connection_string: ODBC Connection string used to connect to the data source. To find a
        connection string for your data source try https://www.connectionstrings.com/.
    :param user: Allows for specifying the user seperatly from the connection string if it is not
        already part of it. The value will eventually be escaped and attached to the connection
        string as `UID`.
    :param password: Allows for specifying the password seperatly from the connection string if it
        is not already part of it. The value will eventually be escaped and attached to the
        connection string as `PWD`.
    :param login_timeout_sec: Number of seconds to wait for a login request to complete before
        returning to the application. The default is driver-dependent. If ``0``, the timeout is
        disabled and a connection attempt will wait indefinitely. If the specified timeout exceeds
        the maximum login timeout in the data source, the driver substitutes that value and uses
        that instead.
    :param packet_size: Specifying the network packet size in bytes. Many ODBC drivers do not
        support this option. If the specified size exceeds the maximum packet size or is smaller
        than the minimum packet size, the driver substitutes that value and returns SQLSTATE 01S02
        (Option value changed).You may want to enable logging to standard error using
        ``log_to_stderr``.
    """
    connection = connect(connection_string, user, password, login_timeout_sec, packet_size)
    connection.insert_into_table(reader=reader, table=table, chunk_size=chunk_size)


def from_table_to_db(
    source: Any,
    target: str,
    connection_string: str,
    chunk_size: int = 1000,
    user: str | None = None,
    password: str | None = None,
    login_timeout_sec: int | None = None,
):
    """
    Reads an arrow table and inserts its contents into a relational table on the database.

    This is a convinience wrapper around ``insert_into_table`` which converts an arrow table into a
    record batch reader for you.

    Example:

    .. code-block:: python

        from arrow_odbc import from_table_to_db
        import pyarrow as pa
        import pandas


        def dataframe_to_table(df):
            table = pa.Table.from_pandas(df)
            from_table_to_db(
                source=table
                connection_string=connection_string,
                user="SA",
                password="My@Test@Password",
                chunk_size=1000,
                table="MyTable",
            )

    :param source: PyArrow table with content to be inserted into the target table on the database.
        Each column of the table must correspond to a column in the target table with identical
        name. Each field in the schema must correspond to a column in the table with identical name.
        The iterator must yield individual arrow tables. In case a column name contains a
        non-alphanumeric character different from ``@``, ``$``, ``#``, or ``_``, the name will be
        escaped using double quotes. However if the name already is already escaped, i.e. it is
        enclosed in either rectangular brackets ([, ]), double quotes (") or backticks (`), it will
        be used as is. This is to allow for users to apply their own quoting if needed, e.g. in case
        reserved keywords are used as column names.
    :param target: Name of the database table to insert into.
    :param connection_string: ODBC Connection string used to connect to the data source. To find a
        connection string for your data source try https://www.connectionstrings.com/.
    :param chunk_size: Number of records to insert in each roundtrip to the database. The number
        will be automatically reduced to the number of rows, if the table is small, in order to save
        memory.
    :param user: Allows for specifying the user seperatly from the connection string if it is not
        already part of it. The value will eventually be escaped and attached to the connection
        string as `UID`.
    :param password: Allows for specifying the password seperatly from the connection string if it
        is not already part of it. The value will eventually be escaped and attached to the
        connection string as `PWD`.
    :param login_timeout_sec: Number of seconds to wait for a login request to complete before
        returning to the application. The default is driver-dependent. If ``0``, the timeout is
        disabled and a connection attempt will wait indefinitely. If the specified timeout exceeds
        the maximum login timeout in the data source, the driver substitutes that value and uses
        that instead.
    """
    connection = connect(
        connection_string=connection_string,
        user=user,
        password=password,
        login_timeout_sec=login_timeout_sec,
    )
    connection.from_table_to_db(
        source=source,
        target=target,
        chunk_size=chunk_size,
    )
